from actionpack.actions import ReadBytes

from oslash import Left
from oslash import Right
from pathlib import Path
from unittest import TestCase
from unittest.mock import patch


class ReadInputTest(TestCase):

    contents = 'some file contents.'.encode()

    @patch('pathlib.Path.read_bytes')
    def test_can_ReadBytes(self, mock_input):
        mock_input.return_value = self.contents
        result = ReadBytes(__file__).perform()
        self.assertIsInstance(result, Right)
        self.assertEqual(result.value, self.contents)

    @patch('pathlib.Path.read_bytes')
    def test_can_ReadBytes_even_if_failure(self, mock_input):
        mock_input.return_value = self.contents

        invalid_file_result = ReadBytes('some/invalid/filepath').perform()
        self.assertIsInstance(invalid_file_result, Left)
        self.assertIsInstance(invalid_file_result.value, FileNotFoundError)

        directory_result = ReadBytes(Path(__file__).parent).perform()
        self.assertIsInstance(directory_result, Left)
        self.assertIsInstance(directory_result.value, IsADirectoryError)




from actionpack.actions import MakeRequest
from actionpack.actions import RetryPolicy
from tests.actionpack.actions import FakeResponse

from oslash import Left
from oslash import Right
from unittest import TestCase
from unittest.mock import Mock
from unittest.mock import patch


class RetryPolicyTest(TestCase):

    @patch('requests.Session.send')
    def test_can_enact_RetryPolicy(self, mock_session_send):
        mock_session_send.side_effect = Exception('something went wrong :/')
        max_retries = 2
        action = RetryPolicy(MakeRequest('GET', 'http://localhost'), max_retries=max_retries)
        result = action.perform()

        self.assertEqual(action.retries, max_retries)
        self.assertIsInstance(result, Left)
        self.assertIsInstance(result.value, RetryPolicy.Expired)

    @patch('requests.Session.send')
    def test_can_enact_RetryPolicy_that_ultimately_succeeds(self, mock_session_send):
        mock_session_send.side_effect = [
            Exception('something went wrong :/'),
            FakeResponse('something went right :)'.encode())
        ]
        max_retries = 2
        action = RetryPolicy(MakeRequest('GET', 'http://localhost'), max_retries=max_retries)
        result = action.perform()

        self.assertEqual(action.retries, 1)
        self.assertIsInstance(result, Right)



from actionpack import Action
from actionpack import KeyedProcedure
from actionpack import Procedure
from tests.actionpack import FakeAction

from collections.abc import Iterable
from oslash import Left
from oslash import Right
from unittest import TestCase


class ProcedureTest(TestCase):

    def assertIsIterable(self, x):
        return isinstance(x, Iterable)

    def setUp(self):
        self.success = FakeAction(name='success')
        self.failure = FakeAction(name='failure',
                                  exception=Exception('something went wrong :/'))
        self.procedure = Procedure(self.success, self.failure)

    def test_Procedure_is_Iterable_of_Actions(self):
        self.assertIsIterable(self.procedure)
        self.assertIsInstance(next(self.procedure), FakeAction)
        self.assertIsInstance(next(self.procedure), FakeAction)
        self.assertIs(next(self.procedure), None)

    def test_Procedure_execution_yields_results(self):
        results = self.procedure.execute()

        self.assertIsIterable(results)
        self.assertIsInstance(next(results), Right)
        self.assertIsInstance(next(results), Left)

    def test_can_validate_Procedure(self):
        with self.assertRaises(Procedure.NotAnAction):
            Procedure('wut.', self.failure).validate()

    # TODO (withtwoemms) -- add concurrency tests

    def test_can_create_KeyedProcedure(self):
        results = KeyedProcedure(self.success, self.failure).execute()
        results_dict = dict(results)

        self.assertIsInstance(results_dict[self.success.name], Right)

    def test_can_validate_KeyedProcedure(self):
        with self.assertRaises(KeyedProcedure.UnnamedAction):
            KeyedProcedure(FakeAction(), self.failure)




from actionpack import Action
from tests.actionpack import FakeAction

from oslash import Left
from oslash import Right
from unittest import TestCase


class ActionTest(TestCase):

    def test_Action_produces_Right_result_when_performed(self):
        result = FakeAction().perform()
        self.assertIsInstance(result, Right)
        self.assertEqual(result.value, FakeAction.result)

    def test_Action_produces_Left_if_exception_raised_when_performed(self):
        exception = Exception('something went wrong :/')
        result = FakeAction(exception=exception).perform()
        self.assertIsInstance(result, Left)
        self.assertEqual(result.value, exception)




from actionpack import Action

from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import as_completed
from functools import reduce
from multiprocessing.pool import ThreadPool
from typing import List


class Procedure:
    def __init__(self, *actions: List[Action], sync: bool=True):
        self.actions = actions
        self._actions = iter(self.actions)
        self.sync = sync

    def validate(self):
        try:
            for action in self.actions:
                if not isinstance(action, Action):
                    msg = f'Procedures can only execute Actions: {str(action)}'
                    raise Procedure.NotAnAction(msg)
        except Exception as e:
            raise e

    def execute(self, max_workers: int=5):
        if self.sync:
            for action in self.actions:
                yield action.perform()
        else:
            with ThreadPoolExecutor(max_workers=max_workers) as executor:
                futures = {executor.submit(action.perform): str(action) for action in self}
                for future in as_completed(futures):
                    yield future.result()

    def __repr__(self):
        for action in self.actions:
            header = f'\nProcedure for {"synchronously" if self.sync else "asynchronously"} performing\nthe following Actions:\n'
            bullet = '  * '
            actions = reduce(lambda a, b: str(a) + f'\n{bullet}' + str(b), self.actions)
            return header + bullet + actions

    def __iter__(self):
        return self._actions

    def __next__(self):
        try:
            return next(self._actions)
        except StopIteration:
            self._actions = iter(self.actions)

    class NotAnAction(Exception): pass


class KeyedProcedure(Procedure):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.validate()

    def validate(self):
        for action in self.actions:
            if not action.__dict__.get('name'):
                msg = f'All {self.__class__.__name__} Actions must have a name: {str(action)}'
                raise KeyedProcedure.UnnamedAction(msg)

    def execute(self, max_workers: int=5):
        if self.sync:
            for action in self.actions:
                yield action.name, action.perform()
        else:
            with ThreadPoolExecutor(max_workers=max_workers) as executor:
                futures = {executor.submit(action.perform): action for action in self}
                for future in as_completed(futures):
                    yield futures[future].name, future.result()

    class UnnamedAction(Exception): pass




from actionpack import Action
from pathlib import Path


class WriteBytes(Action):
    def __init__(self, filename: str, bytes_to_write: bytes, overwrite: bool=False):
        self.path = Path(filename)
        self.overwrite = overwrite
        self.instruction = lambda: self.path.write_bytes(bytes_to_write)

    def validate(self):
        if self.path.exists() and not self.overwrite:
            raise FileExistsError(f'Cannot {str(self)} to {str(self.path)}')
        if self.path.is_dir():
            raise IsADirectoryError(str(self.path))
        return self




from actionpack.utils import tally

from unittest import TestCase


class UtilsTest(TestCase):

    def test_tally(self):
        self.assertEqual(list(tally(0)), [])
        self.assertEqual(list(tally()), [1])
        self.assertEqual(list(tally(2)), [1, 1])




from actionpack.actions import ReadInput

from unittest import TestCase
from unittest.mock import patch


class ReadInputTest(TestCase):

    @patch('builtins.input')
    def test_can_ReadInput(self, mock_input):
        reply = 'I\'m fine.'
        mock_input.return_value = reply
        action = ReadInput('How are you?')
        self.assertEqual(action.perform().value, reply)

